from graphql_relay.node.node import (
    to_global_id
)
from graphql_relay.connection.connection import (
    connection_definitions
)

from graphene.core.types import Interface, ObjectType
from graphene.core.fields import BooleanField, StringField, ListField, Field
from graphene.relay.fields import GlobalIDField
from graphene.utils import memoize


class BaseNode(object):
    @classmethod
    @memoize
    def get_connection(cls, schema):
        _type = cls.internal_type(schema)
        type_name = cls._meta.type_name
        connection = connection_definitions(type_name, _type).connection_type
        return connection

    @classmethod
    def _prepare_class(cls):
        from graphene.relay.utils import is_node
        if is_node(cls):
            assert hasattr(
                cls, 'get_node'), 'get_node classmethod not found in %s Node' % cls

    @classmethod
    def to_global_id(cls, instance, args, info):
        type_name = cls._meta.type_name
        return to_global_id(type_name, instance.id)


class Node(BaseNode, Interface):
    '''An object with an ID'''
    id = GlobalIDField()


class PageInfo(ObjectType):
    has_next_page = BooleanField(required=True, description='When paginating forwards, are there more items?')
    has_previous_page = BooleanField(required=True, description='When paginating backwards, are there more items?')
    start_cursor = StringField(description='When paginating backwards, the cursor to continue.')
    end_cursor = StringField(description='When paginating forwards, the cursor to continue.')


class Edge(ObjectType):
    '''An edge in a connection.'''
    node = Field(lambda field: field.object_type.node_type, description='The item at the end of the edge')
    end_cursor = StringField(required=True, description='A cursor for use in pagination')

    @classmethod
    @memoize
    def for_node(cls, node):
        from graphene.relay.utils import is_node
        assert is_node(node), 'ObjectTypes in a edge have to be Nodes'
        return type('%sEdge' % node._meta.type_name, (cls, ), {'node_type': node})


class Connection(ObjectType):
    '''A connection to a list of items.'''
    page_info = Field(PageInfo, required=True, description='The Information to aid in pagination')
    edges = ListField(lambda field: field.object_type.edge_type, description='Information to aid in pagination.')

    @classmethod
    @memoize
    def for_node(cls, node, edge_type=None):
        from graphene.relay.utils import is_node
        edge_type = edge_type or Edge
        assert is_node(node), 'ObjectTypes in a connection have to be Nodes'
        return type('%sConnection' % node._meta.type_name, (cls, ), {'edge_type': edge_type.for_node(node)})



from graphene.relay.fields import (
    ConnectionField,
    NodeField,
    GlobalIDField,
)

from graphene.relay.types import (
	Node,
	PageInfo,
	Edge,
	Connection
)

from graphene.relay.utils import is_node



from collections import Iterable, OrderedDict

from graphql_relay.connection.arrayconnection import (
    connection_from_list
)
from graphql_relay.connection.connection import (
    connectionArgs
)
from graphql_relay.node.node import (
    from_global_id
)
from graphql.core.type import (
    GraphQLNonNull,
    GraphQLID,
    GraphQLArgument,
)

from graphene.core.fields import Field, IDField
from graphene.utils import memoize


class ConnectionField(Field):

    def __init__(self, field_type, resolve=None, description='', connection_type=None, edge_type=None, **kwargs):
        from graphene.relay.types import Connection, Edge
        super(ConnectionField, self).__init__(field_type, resolve=resolve,
                                              args=connectionArgs, description=description, **kwargs)
        self.connection_type = connection_type or Connection
        self.edge_type = edge_type or Edge
        assert issubclass(self.connection_type, Connection), 'connection_type in %r must be a subclass of Connection' % self
        assert issubclass(self.edge_type, Edge), 'edge_type in %r must be a subclass of Edge' % self

    def wrap_resolved(self, value, instance, args, info):
        return value

    def resolve(self, instance, args, info):
        resolved = super(ConnectionField, self).resolve(instance, args, info)
        if resolved:
            resolved = self.wrap_resolved(resolved, instance, args, info)
            assert isinstance(
                resolved, Iterable), 'Resolved value from the connection field have to be iterable'
            return connection_from_list(resolved, args)

    @memoize
    def internal_type(self, schema):
        from graphene.relay.utils import is_node
        node = self.get_object_type(schema)
        assert is_node(node), 'Only nodes have connections.'
        schema.register(node)
        edge_node_type = self.edge_type.for_node(node)
        connection_node_type = self.connection_type.for_node(node, edge_type=edge_node_type)
        return connection_node_type.internal_type(schema)


class NodeField(Field):
    def __init__(self, object_type=None, *args, **kwargs):
        from graphene.relay.types import Node
        super(NodeField, self).__init__(object_type or Node, *args, **kwargs)
        self.field_object_type = object_type
        self.args['id'] = GraphQLArgument(
            GraphQLNonNull(GraphQLID),
            description='The ID of an object'
        )

    def id_fetcher(self, global_id, info):
        from graphene.relay.utils import is_node
        schema = info.schema.graphene_schema
        resolved_global_id = from_global_id(global_id)
        _type, _id = resolved_global_id.type, resolved_global_id.id
        object_type = schema.get_type(_type)
        if not is_node(object_type) or (self.field_object_type and
           object_type != self.field_object_type):
            return

        return object_type.get_node(_id)

    def resolve(self, instance, args, info):
        global_id = args.get('id')
        return self.id_fetcher(global_id, info)


class GlobalIDField(IDField):
    '''The ID of an object'''
    required = True

    def contribute_to_class(self, cls, name, add=True):
        from graphene.relay.utils import is_node, is_node_type
        in_node = is_node(cls) or is_node_type(cls)
        assert in_node, 'GlobalIDField could only be inside a Node, but got %r' % cls
        super(GlobalIDField, self).contribute_to_class(cls, name, add)

    def resolve(self, instance, args, info):
        return self.object_type.to_global_id(instance, args, info)



from graphene.core.fields import (
    ListField
)
from graphene import relay

from graphene.core.fields import Field, LazyField
from graphene.utils import cached_property, memoize, LazyMap

from graphene.relay.types import BaseNode
from graphene.relay.utils import is_node
from graphene.contrib.django.utils import get_type_for_model, lazy_map


class DjangoConnectionField(relay.ConnectionField):
    def wrap_resolved(self, value, instance, args, info):
        schema = info.schema.graphene_schema
        return lazy_map(value, self.get_object_type(schema))


class LazyListField(ListField):
    def resolve(self, instance, args, info):
        schema = info.schema.graphene_schema
        resolved = super(LazyListField, self).resolve(instance, args, info)
        return lazy_map(resolved, self.get_object_type(schema))


class ConnectionOrListField(LazyField):
    @memoize
    def get_field(self, schema):
        model_field = self.field_type
        field_object_type = model_field.get_object_type(schema)
        if is_node(field_object_type):
            field = DjangoConnectionField(model_field)
        else:
            field = LazyListField(model_field)
        field.contribute_to_class(self.object_type, self.name)
        return field


class DjangoModelField(Field):
    def __init__(self, model, *args, **kwargs):
        super(DjangoModelField, self).__init__(None, *args, **kwargs)
        self.model = model

    def resolve(self, instance, args, info):
        resolved = super(DjangoModelField, self).resolve(instance, args, info)
        schema = info.schema.graphene_schema
        _type = self.get_object_type(schema)
        assert _type, ("Field %s cannot be retrieved as the "
                       "ObjectType is not registered by the schema" % (
                           self.field_name
                       ))
        return _type(resolved)

    @memoize
    def internal_type(self, schema):
        _type = self.get_object_type(schema)
        if not _type and self.object_type._meta.only_fields:
            raise Exception(
                "Model %r is not accessible by the schema. "
                "You can either register the type manually "
                "using @schema.register. "
                "Or disable the field %s in %s" % (
                    self.model,
                    self.field_name,
                    self.object_type
                )
            )
        return _type and _type.internal_type(schema) or Field.SKIP

    def get_object_type(self, schema):
        return get_type_for_model(schema, self.model)
